// Copyright (C) 2013 S&P Capital IQ
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.spcapitaliq.platformexecuting

import java.io.{BufferedOutputStream, File, FileOutputStream,
                InputStream, IOException}
import java.net.URI
import java.util.zip.{ZipEntry, ZipFile, ZipInputStream}
import scala.annotation.tailrec

import scalaz.{@@, \/, NonEmptyList, Tag}
import scalaz.std.string._
import scalaz.syntax.functor._
import scalaz.syntax.id._
import scalaz.syntax.semigroup._
import scalaz.syntax.std.boolean._
import scalaz.syntax.std.list._
import scalaz.syntax.std.option._

import Util.{copyIO, OpenAs}
import ZipEntName._

/** An entry at some arbitrary jar depth.
  *
  * @param extract URI of the jar/zip file itself.
  * @param out Entry names in order of extraction.  To get an entry
  *            trivially, take the `out.head` entry from `extract`,
  *            the `out.tail.head` entry from that; the final entry is
  *            the referred file.
  */
final case class JarURI(extract: URI, out: NonEmptyList[String @@ ZipEntName]) {
  @tailrec
  private[this] def findZipEntry[A](i: ZipInputStream, f: ZipEntry => Option[A]): Option[A] =
    Option(i.getNextEntry()) match {
      case None => None
      case Some(ze) => f(ze) match {
        case None => findZipEntry(i, f)
        case a@Some(_) => a
      }
    }

  /** Assume `out` represents a single file and `f` doesn't cast its
    * argument, and produce either a read error or `f`'s result.
    */
  def withOpenFile[A](f: InputStream => A): IOException \/ A = {
    def rec(is: InputStream, pt: List[String @@ ZipEntName]): IOException \/ A =
      pt match {
        case Nil => f(is).right
        case e :: es =>
          new ZipInputStream(is) openAs {zi =>
            for {
              _ <- findZipEntry(zi, ze => (ZipEntName.from(ze) == e) option (()))
            } yield rec(zi, es)
          } getOrElse (new IOException("No entry " |+| e).left)
      }
    try {
      if (Option(extract.getScheme) cata ("file" equalsIgnoreCase, false)) {
        new ZipFile(new File(extract)) openAs (z =>
          Option(z getEntry out.head)
            cata (e => z getInputStream e openAs (rec(_, out.tail)),
                  new IOException("No entry " |+| out.head).left))
      } else extract.toURL.openStream() openAs (rec(_, out.list))
    } catch {case e: IOException => e.left}
  }

  @tailrec
  private[this] def foldZipEntries[A](i: ZipInputStream, a: A)(f: (A, ZipEntry) => A): A =
    Option(i.getNextEntry()) match {
      case None => a
      case Some(ze) => foldZipEntries(i, f(a, ze))(f)
    }

  /** Assume `out` represents a directory and `f` strictly reads, and
    * produce either a read error or the collected `f` results.
    */
  def foldOpenDirectory[A](a: A)(f: (A, ZipEntry, InputStream) => A): IOException \/ A = {
    def rec(is: InputStream,
            pt: NonEmptyList[String @@ ZipEntName]): IOException \/ A =
      (pt.head, pt.tail.toNel) match {
        case (e, None) =>
          val esl = (e:String) |+| (if (e.isEmpty || (e endsWith "/")) "" else "/")
          new ZipInputStream(is) openAs {zi =>
            foldZipEntries(zi, a)((a, ze) =>
              if (ZipEntName.from(ze) startsWith esl) f(a, ze, zi)
              else a)
          } right
        case (e, Some(es)) =>
          new ZipInputStream(is) openAs {zi =>
            for {
              _ <- findZipEntry(zi, ze => (ZipEntName.from(ze) == e) option (()))
            } yield rec(zi, es)
          } getOrElse (new IOException("No entry " |+| e).left)
      }
    try extract.toURL.openStream() openAs (rec(_, out))
    catch {case e: IOException => e.left}
  }

  /** Assume `out` represents a directory; extract its contents to
    * `d`.  Answer the files & directories created.
    */
  def extractDirectoryTo(d: File): IOException \/ Set[File] =
    foldOpenDirectory(Set.empty[File]){(fcs, ze, is) =>
      val (sd, bn) = zipEntSplit(relativizeZipEnt(last, ZipEntName from ze))
      val dn = new File(d, zipEntToNative(sd))
      dn.mkdirs()
      fcs + (if (bn.isEmpty) dn else {
        val fil = new File(dn, bn)
        new FileOutputStream(fil) openAs {fo =>
          val bo = new BufferedOutputStream(fo)
          copyIO(is, bo)
          bo.flush()
        }
        fil
      })
    }

  /** The last filename component. */
  def baseName: String @@ ZipEntName = zipEntBaseName(last)

  private def last: String @@ ZipEntName = out.tail.lastOption getOrElse out.head
}

/** Parsing Jar URIs and extracting them to disk. */
object Jar {
  private[this] val jaru = """(?s)^(.*?:.*)!/(.*)$""".r

  private[this] def whileDoP[A](a: A)(f: A => Option[A]): A = {
    @tailrec def rec(a: A): A = f(a) match {
      case None => a
      case Some(a2) => rec(a2)
    }
    rec(a)
  }

  /** Parse a URI into jar levels. */
  def jarURI(u: URI): Option[JarURI] = for {
    (u, o) <- Some(whileDoP((u, List.empty[String])){case (u, o) => for {
        scm <- Option(u.getScheme)
        if "jar" equalsIgnoreCase scm
        m <- jaru findFirstMatchIn u.getRawSchemeSpecificPart
         } yield (new URI(m group 1), (m group 2) :: o)})
    uscm <- Option(u.getScheme)
    if !("jar" equalsIgnoreCase uscm)
    o <- o.toNel
  } yield JarURI(u, Tag subst o)

  /** Commonalities of `extractTemp` and `extractTempDir`. */
  private[this]
  def extractTempR[A](u: URI)(f: JarURI => (String \/ (Option[A], File))
                    ): String \/ (Option[A], File) =
    (Option(u.getScheme) \/> ("Malformed URI: " |+| u.toString)) flatMap {
      case scm if scm equalsIgnoreCase "file" => (None, new File(u)).right
      case scm =>
        (jarURI(u).filter(ju =>
           Option(ju.extract.getScheme) cata ("file" equalsIgnoreCase, false))
           \/> (scm |+| " can't be extracted; must be jar or file"))
        .flatMap(f)
    }

  /** Extract `u`, if possible.  `#extract` may be simpler for you.
    *
    * @param u A file, indicated by `file` or `jar` URI.
    * @return An error message or the extracted file, and if `str` was
    *         used to acquire a temporary file, its `#release`
    *         identifier. */
  def extractTemp(u: URI)(implicit str: TempFileStrategy): String \/ (Option[str.Fid], File) =
    extractTempR(u)(ju => ju.withOpenFile{i =>
                      val (st, tmp) = str.acquire(ju.baseName)
                      new FileOutputStream(tmp) openAs (copyIO(i, _))
                      (Some(st), tmp)
                    } leftMap (_.getMessage))

  /** Extract `u`, if possible, and invoke `f` with it.  The extraction
    * is only valid in `f`'s dynamic context.
    */
  def extract[A](u: URI)(f: File => A)(implicit str: TempFileStrategy): String \/ A =
    extractTemp(u) map {case (ofid, fil) =>
        try f(fil)
        finally {ofid foreach str.release}
    }

  /** Extract `u`, if possible.  `#extractDirectory` may be simpler
    * for you.
    *
    * @param u A directory, indicated by `file` or `jar` URI.
    * @return An error message or the extracted directory, and if
    *         `str` was used to acquire a temporary directory, its
    *         `#release` identifier. */
  def extractTempDir(u: URI)(implicit str: TempDirStrategy): String \/ (Option[str.Fid], File) =
    extractTempR(u){ju =>
      val (st, tmp) = str.acquire(ju.baseName)
      (ju extractDirectoryTo tmp
         leftMap (_.getMessage) as ((Some(st), tmp)))
    }

  /** Extract the directory `u`, if possible, and invoke `f` with it.
    * The extraction is only valid in `f`'s dynamic context.
    */
  def extractDirectory[A](u: URI)(f: File => A)(implicit str: TempDirStrategy): String \/ A =
    extractTempDir(u) map {case (ofid, fil) =>
        try f(fil)
        finally {ofid foreach str.release}
    }
}
